package catala

  1. Overview
  2. Docs
Legend:
Page
Library
Module
Module type
Parameter
Class
Class type
Module Clerk_utilsSource

Sourcemodule Diff : sig ... end

A simple diffing algorithm

Sourcemodule Module_graph : sig ... end
Sourcemodule Poll : sig ... end

This module scans the system for the location of various required tools and libraries, to be used as default values. (think ./configure)

Sourcemodule Scan : sig ... end

This module is responsible for scanning Catala files, extracting dependency and test information. It is based on the lightweight "line-parser" (Surface.Parser_driver.line)

Sourcemodule Var : sig ... end