package catala

  1. Overview
  2. Docs
Legend:
Page
Library
Module
Module type
Parameter
Class
Class type
Module Clerk_utils.PollSource

This module scans the system for the location of various required tools and libraries, to be used as default values. (think ./configure)

This module is sensitive to the CWD at first use. Therefore it's expected that chdir has been run beforehand to the project root.

Locates the main lib directory containing the OCaml libs

Locates the directory containing the runtimes to link to (in one subdirectory per backend)

Sourceval catala_source_tree_root : Catala_utils.File.t option Lazy.t

Set if, and only if the current build happens from within the Catala compiler source tree itself. This requires some specific options and includes for correct bootstrapping, and running the tests without requiring an installation first.