package catala

  1. Overview
  2. Docs
Legend:
Page
Library
Module
Module type
Parameter
Class
Class type
Module Var.OpSource

Sourceval (!) : string t -> string

Alias to Var.ref. Run-time reference to the given variable !var = "${xvarname}", for inclusion in strings

Sourceval (!!) : 'a t -> Ninja_utils.Expr.elt

Run-time reference to the given variable !var = "${xvarname}" as an expression element that will be appropriately quoted. This handles specific quoting rules of the input and output variables by Ninja