package catala

  1. Overview
  2. Docs
Legend:
Page
Library
Module
Module type
Parameter
Class
Class type
Module Desugared.PrintSource

Sourceval exceptions_graph : is_condition:bool -> Shared_ast.ScopeName.t -> Ast.ScopeDef.t -> Dependency.ExceptionsDependencies.t -> unit

Prints the exception graph of a variable to the terminal

Sourceval exceptions_graph_json : is_condition:bool -> Shared_ast.ScopeName.t -> Ast.ScopeDef.t -> Dependency.ExceptionsDependencies.t -> unit

Prints the exception graph of a variable as JSON to stdout

Builds the exception tree from the graph

Sourceval exception_tree_to_json : exception_tree -> Yojson.Safe.t

Serialises an exception tree node to a JSON value