package catala

  1. Overview
  2. Docs
Legend:
Page
Library
Module
Module type
Parameter
Class
Class type
Module Catala_runtime.PrintSource

This module is for setting options and internals, use Value.format to print values

Sourcetype lang = [
  1. | `En
  2. | `Fr
  3. | `Pl
]
Sourceval set_lang : [ `En | `Fr | `Pl ] -> unit

Sets the language to be used for the output of values

Sourceval get_lang : unit -> [ `En | `Fr | `Pl ]
Sourceval set_precision : int -> unit

Sets the maximum number of decimal numbers to print

Sourceval get_precision : unit -> int